
//#define GV_SAMPLER_2D_SHADOW


// --- SHADOWS ---
#ifdef GV_SAMPLER_2D_SHADOW
	uniform sampler2DShadow shadowMap;
#else
	uniform sampler2D shadowMap;
#endif

uniform bool      usePCF;
// ---


// --- LIGHTING ---
uniform sampler2D texture;
uniform bool      hasTexture;
// ---



varying vec3 normal, lightVector, halfVector;
varying vec4 shadowCoord, posLight;



#ifdef GV_SAMPLER_2D_SHADOW

float shadowOB( float zValue, vec4 shadowCoordH, float epsilonOffset )
{
	//return shadow2DProj( shadowMap, shadowCoordH ).x;
	return shadow2DProj( shadowMap, shadowCoordH + epsilonOffset ).x;
}

#else

/*
 *  Calculate simple shadows
 */
float shadow( float zValue, vec4 shadowCoordH, float epsilonOffset )
{
	//float shadowValue = texture2DProj( shadowMap, shadowCoordH + epsilonOffset ).x; // ev. texture2DProj(...)
	//float shadowValue = texture2D( shadowMap, shadowCoordH.st + 0.5/256.0 ).r + epsilonOffset; // ev. texture2DProj(...)
	float shadowValue = texture2D( shadowMap, shadowCoordH.st ).r + epsilonOffset; // ev. texture2DProj(...)

	float lightAmmount = 0.0;
	if ( shadowValue < zValue ) {
		// we are in shadow
		lightAmmount = 0.0;
	} else {
		// we are in light
		lightAmmount = 1.0;
	}

	return lightAmmount;
}


/*
 *  Calculate shadows using PCF
 */
float shadowPCF( float zValue, vec4 shadowCoordH, float epsilonOffset, float textureSize )
{
	vec2  shadowCoords[4];
	float shadowValues[4];

	// for PCF
	shadowCoords[0] = shadowCoordH.st;
	shadowCoords[1] = shadowCoordH.st + vec2( 1.0/textureSize,             0.0 );
	shadowCoords[2] = shadowCoordH.st + vec2( 0.0,             1.0/textureSize );
	shadowCoords[3] = shadowCoordH.st + vec2( 1.0/textureSize, 1.0/textureSize );
/*
	epsilonOffset = 0.001;

	vec4  shadowCoords[4];
	float shadowValues[4];

	// for PCF
	shadowCoords[0] = shadowCoordH;
	shadowCoords[1] = shadowCoordH + vec4( 1.0/textureSize,             0.0, 0.0, 0.0 );
	shadowCoords[2] = shadowCoordH + vec4( 0.0,             1.0/textureSize, 0.0, 0.0 );
	shadowCoords[3] = shadowCoordH + vec4( 1.0/textureSize, 1.0/textureSize, 0.0, 0.0 );
*/	

	for ( int i = 0; i < 4; ++i ) {
		//shadowValues[i] = texture2DProj( shadowMap, shadowCoords[i] + epsilonOffset ).r; // ev. texture2DProj(...)
		shadowValues[i] = texture2D( shadowMap, shadowCoords[i].st ).r + epsilonOffset; // ev. texture2DProj(...)
		shadowValues[i] = shadowValues[i] < zValue ? 0.0 : 1.0;
	}

	vec4  iPolFactor       = fract( shadowCoordH * textureSize );
	float finalShadowValue = mix( mix( shadowValues[0], shadowValues[1], iPolFactor.x ),
								  mix( shadowValues[2], shadowValues[3], iPolFactor.x ),
								  iPolFactor.y 
								 );

	return finalShadowValue;
}
#endif



/*
 *  Calculate per-pixel lighting
 */
vec4 calculateLighting( float lightAmmount )
{
	// --- calculate lighting ---
    vec3  diffuse   = gl_FrontMaterial.diffuse.rgb * gl_LightSource[0].diffuse .rgb;
    vec3  ambient   = gl_FrontMaterial.ambient.rgb * ( gl_LightSource[0].ambient.rgb + gl_LightModel.ambient.rgb );

    vec3  specular  = gl_FrontMaterial.specular.rgb * gl_LightSource[0].specular.rgb;
    float shininess = gl_FrontMaterial.shininess;
    
	vec3 n = normalize( normal );
    float NdotL = max( dot( n, lightVector ), 0.0 );
    
	diffuse  *= lightAmmount;
	specular *= lightAmmount;

	vec3 d;
	if ( hasTexture ) {
		vec3 tex = texture2D( texture, gl_TexCoord[0].st ).rgb;
		d = tex * ( diffuse * NdotL + ambient );
	} else {
	    d = diffuse * NdotL + ambient;
	}

    vec3 s = specular * pow( max( dot( n, normalize( halfVector )), 0.0 ), shininess );
	vec4 fragColor = vec4( min( d + s, 1.0), 1.0 );
	
	return fragColor;
}




/*  -------------
 *  Fragment-MAIN
 *  -------------
 */
void main()
{
	// --- SHADOWS ---
	float epsilonOffset = 0.005;  // 0.005; // 0.01; // 0.001; // was: 0.0005
	float textureSize   = 2048.0;
	vec4  shadowCoordH  = shadowCoord / shadowCoord.w;
	float zValue        = posLight.z / posLight.w;
	float lightAmmount  = 1.0;


	// --- calculate shadow	---
#ifdef GV_SAMPLER_2D_SHADOW
	lightAmmount = shadowOB( zValue, shadowCoordH, epsilonOffset );
#else
	if ( usePCF ) lightAmmount = shadowPCF( zValue, shadowCoordH, epsilonOffset, textureSize );
	else          lightAmmount = shadow( zValue, shadowCoordH, epsilonOffset );
#endif


	// --- calculate lighting ---
	vec4 fragColor = calculateLighting( lightAmmount );
    gl_FragColor = fragColor;
}
